//
//  PluginHandler.swift
//  Pods-Runner
//
//  Created by Austin Goh on 21/04/2020.
//

import Foundation
import Flutter
import TwilioChatClient

public class PluginHandler {
    
    private let CLIENT_IS_NULL_CODE = "CLIENT_IS_NULL"
    private let CLIENT_IS_NULL_MESSAGE = "Client has not been initialized"
    
    private let CHANNEL_IS_NULL_CODE = "CHANNEL_IS_NULL"
    private let CHANNEL_IS_NULL_MESSAGE = "Channel has not been initialized"
    
    private let USER_IS_NULL_CODE = "USER_IS_NULL"
    private let USER_IS_NULL_MESSAGE = "User has not been initialized"
    
    private let ARGUMENT_IS_NOT_NOT_PROVIDED_CODE = "ARGUMENT_IS_NOT_NOT_PROVIDED"
    private let ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE = "One of the arguments is missing"
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        SwiftTwilioChatPlugin.debug("PluginHandler.handle => received \(call.method)")
        switch call.method {
        case "debug":
            debug(call, result: result)
        case "createChatClient":
            createChatClient(call, result: result)
        case "updateUserFriendlyName":
            updateUserFriendlyName(call, result: result)
        case "registerFCMToken":
            registerFCMToken(call, result: result)
        case "unregisterFCMToken":
            unregisterFCMToken(call, result: result)
        case "replaceFCMToken":
            replaceFCMToken(call, result: result)
        case "enterChannel":
            enterChannel(call, result: result)
        case "exitChannel":
            exitChannel(call, result: result)
        case "listPublicChannels":
            listPublicChannels(call, result: result)
        case "listUserChannels":
            listUserChannels(call, result: result)
        case "createChannel":
            createChannel(call, result: result)
        case "joinChannel":
            joinChannel(call, result: result)
        case "leaveChannel":
            leaveChannel(call, result: result)
        case "sendMessagesByChannelSID":
            sendMessagesByChannelSID(call, result: result)
        case "getMostRecentMessagesByChannelSID":
            getMostRecentMessagesByChannelSID(call, result: result)
        case "inviteUserToChannel":
            inviteUserToChannel(call, result: result)
        case "updateChannelFriendlyName":
            updateChannelFriendlyName(call, result: result)
        default:
            result(FlutterMethodNotImplemented)
        }
    }
    
    private func debug(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'native' parameter", details: nil))
        }
        
        guard let enableNative = arguments["native"] as? Bool else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'native' parameter", details: nil))
        }
        SwiftTwilioChatPlugin.debug("PluginHandler.handle => received \(call.method)")
        SwiftTwilioChatPlugin.nativeDebug = enableNative
        result(enableNative)
    }
    
    func application(_ application: UIApplication, didRegisterForRemoteNotificationsWithDeviceToken deviceToken: Data) {
        SwiftTwilioChatPlugin.deviceToken = deviceToken
        self.registerDeviceToken(deviceToken: deviceToken)
    }
    
    func application(_ application: UIApplication, didFailToRegisterForRemoteNotificationsWithError error: Error) {
        SwiftTwilioChatPlugin.debug("Failed to get token, error: "+error.localizedDescription)
        SwiftTwilioChatPlugin.deviceToken = nil
    }
    
    private func createChatClient(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let accessToken = arguments["accessToken"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'acessToken' parameter", details: nil))
        }
        if SwiftTwilioChatPlugin.chatClient == nil {
            TwilioChatClient.chatClient(withToken: accessToken, properties: nil, delegate: SwiftTwilioChatPlugin.chatClientListener) {
                res, chatClient in
                guard (res.isSuccessful()) else {
                    SwiftTwilioChatPlugin.debug("Failed -> \(res.description)")
                    return result(
                        FlutterError(code: "\(String(describing: res.error?.code))",
                            message: res.error?.description,
                            details: res.error.debugDescription))
                }
                SwiftTwilioChatPlugin.chatClientListener = ChatClientListener()
                chatClient?.delegate = SwiftTwilioChatPlugin.chatClientListener
                SwiftTwilioChatPlugin.connected = true
                SwiftTwilioChatPlugin.chatClient = chatClient
                result(SerializationHelper.chatClientToDict(chatClient!))
                
            }
            
        } else {
            result(SerializationHelper.chatClientToDict(SwiftTwilioChatPlugin.chatClient!))
        }
    }
    
    private func updateUserFriendlyName(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let friendlyName = arguments["friendlyName"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'friendlyName' parameter", details: nil))
        }
        myChatClient.user?.setFriendlyName(friendlyName){
            res in
            guard (res.isSuccessful()) else {
                SwiftTwilioChatPlugin.debug("Failed -> \(res.description)")
                return result(
                    FlutterError(code: "\(String(describing: res.error?.code))",
                        message: res.error?.description,
                        details: res.error.debugDescription))
            }
            return result(true)
            
        }
    }
    
    private func registerFCMToken(_ call: FlutterMethodCall, result: @escaping FlutterResult){}
    
    private func unregisterFCMToken(_ call: FlutterMethodCall, result: @escaping FlutterResult){}
    
    private func replaceFCMToken(_ call: FlutterMethodCall, result: @escaping FlutterResult){}
    
    private func registerDeviceToken(deviceToken: Data) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return;
        }
        guard myChatClient.user != nil else {
            return;
        }
        myChatClient.register(withNotificationToken: deviceToken) { (result) in
            if (!result.isSuccessful()) {
                print("Token is registered")
            }
        }
    }
    
    private func enterChannel(_ call: FlutterMethodCall, result: @escaping FlutterResult){
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, message: ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, details: nil))
        }
        
        guard let channelSID = arguments["channelSID"] as? String else {
            return result(FlutterError(code: ARGUMENT_IS_NOT_NOT_PROVIDED_CODE, message: ARGUMENT_IS_NOT_NOT_PROVIDED_MESSAGE, details: nil))
        }
        
        SwiftTwilioChatPlugin.chatClient?.channelsList()?.channel(withSidOrUniqueName: channelSID){
            res, channel in
            guard (res.isSuccessful()) else {
                SwiftTwilioChatPlugin.debug("Failed -> \(res.description)")
                return result(
                    FlutterError(code: "\(String(describing: res.error?.code))",
                        message: res.error?.description,
                        details: res.error.debugDescription))
            }
            channel?.members?.members(){
                membersRes, membersList in
                var resMembersList = [[String: Any]]()
                if(!membersRes.isSuccessful()){
                    return result(FlutterError(code: "ERROR_GETTING_MEMBERS", message: "Cannot get members from channel", details: nil))
                }
                if(membersList != nil){
                    
                    for member in (membersList?.items())!{
                        resMembersList.append(SerializationHelper.memberToDict(member))
                    }
                }
                SwiftTwilioChatPlugin.channelListener = ChannelListener()
                
                channel?.delegate = SwiftTwilioChatPlugin.channelListener
                SwiftTwilioChatPlugin.channel = channel
                result(SerializationHelper.channelToDict(channel!, resMembersList))
            }
            
            
        }
        
    }
    
    private func exitChannel(_ call: FlutterMethodCall, result: @escaping FlutterResult){
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        SwiftTwilioChatPlugin.channelListener = nil;
        SwiftTwilioChatPlugin.channel = nil
        result(true)
    }
    
    private func listPublicChannels(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        SwiftTwilioChatPlugin.channelsList = myChatClient.channelsList()
        PluginHandler.populateChannels()
        SwiftTwilioChatPlugin.channelsList?.publicChannelDescriptors { res, paginator in
            var channelDescriptorsResult = [[String: Any]]()
            for ch in paginator!.items() {
                channelDescriptorsResult.append(SerializationHelper.channelDescriptorToDict(ch))
            }
            result(channelDescriptorsResult)
        }
    }
    
    private func listUserChannels(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        
        SwiftTwilioChatPlugin.channelsList = myChatClient.channelsList()
        PluginHandler.populateChannels()
        
        SwiftTwilioChatPlugin.channelsList?.userChannelDescriptors { res, paginator in
            var channelDescriptorsResult = [[String: Any]]()
            for ch in paginator!.items() {
                channelDescriptorsResult.append(SerializationHelper.channelDescriptorToDict(ch))
            }
            result(channelDescriptorsResult)
        }
        
    }
    
    private func createChannel(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let channelName = arguments["channelName"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelName' parameter", details: nil))
        }
        guard let channelType = arguments["channelType"] as? Int else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelType' parameter", details: nil))
        }
        
        let channelOptions = [
            TCHChannelOptionFriendlyName: channelName,
            TCHChannelOptionType: channelType
            ] as [String : Any]
        myChatClient.channelsList()?.createChannel(options: channelOptions){ res, channel in
            result(channel)
        }
    }
    
    private func joinChannel(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let channelSID = arguments["channelSID"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelSID' parameter", details: nil))
        }
        
        myChatClient.channelsList()?.channel(withSidOrUniqueName: channelSID) {
            res, channel in
            channel?.join() {
                joinRes in
                if(!joinRes.isSuccessful()){
                    return result(FlutterError(code: "ERROR_GETTING_CHANNEL", message: "Cannot join channel", details: nil))
                }
                channel?.members?.members(){
                    membersRes, membersList in
                    var resMembersList = [[String: Any]]()
                    if(!membersRes.isSuccessful()){
                        return result(FlutterError(code: "ERROR_GETTING_MEMBERS", message: "Cannot get members from channel", details: nil))
                    }
                    if(membersList != nil){
                        for member in (membersList?.items())!{
                            resMembersList.append(SerializationHelper.memberToDict(member))
                        }
                    }
                    result(SerializationHelper.channelToDict(channel!, resMembersList))
                }
            }
        }
    }
    
    private func leaveChannel(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let channelSID = arguments["channelSID"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelSID' parameter", details: nil))
        }
        
        myChatClient.channelsList()?.channel(withSidOrUniqueName: channelSID) {
            res, channel in
            channel?.leave(){
                joinRes in
                result(SerializationHelper.channelToDict(channel!, [[String: Any]]()))
            }
        }
    }
    
    private func sendMessagesByChannelSID(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let channelSID = arguments["channelSID"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelSID' parameter", details: nil))
        }
        
        guard let messageBody = arguments["messageBody"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'messageBody' parameter", details: nil))
        }
        
        let messageOptions = TCHMessageOptions()
        messageOptions.withBody(messageBody)
        myChatClient.channelsList()?.channel(withSidOrUniqueName: channelSID) {
            channelRes, channel in
            channel?.messages?.sendMessage(with:messageOptions){
                sendMsgRes, message in
                result(SerializationHelper.messageToDict(message!))
            }
        }
        
    }
    
    private func getMostRecentMessagesByChannelSID(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        
        guard let myChatClient = SwiftTwilioChatPlugin.chatClient else {
            return result(FlutterError(code: CLIENT_IS_NULL_CODE, message: CLIENT_IS_NULL_MESSAGE, details: nil))
        }
        guard myChatClient.user != nil else {
            return result(FlutterError(code: USER_IS_NULL_CODE, message: USER_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let channelSID = arguments["channelSID"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelSID' parameter", details: nil))
        }
        
        guard let messageCount = arguments["messageCount"] as? Int else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'messageCount' parameter", details: nil))
        }
        
        myChatClient.channelsList()?.channel(withSidOrUniqueName: channelSID) {
            channelRes, channel in
            channel?.messages?.getLastWithCount(UInt(messageCount)){
                getMessageRes, messagesList in
                var messagesResult = [[String: Any]]()
                for msg in messagesList ?? [TCHMessage]() {
                    messagesResult.append(SerializationHelper.messageToDict(msg))
                }
                result(messagesResult)
            }
        }
        
    }
    
    private func inviteUserToChannel(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let channel = SwiftTwilioChatPlugin.channel else {
            return result(FlutterError(code: CHANNEL_IS_NULL_CODE, message: CHANNEL_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        guard let identity = arguments["identity"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'identity' parameter", details: nil))
        }
        guard let channelSID = arguments["channelSID"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelSID' parameter", details: nil))
        }
        if(channelSID == channel.sid){
            channel.members?.invite(byIdentity: identity){
                res in
                guard (res.isSuccessful()) else {
                    SwiftTwilioChatPlugin.debug("Failed -> \(res.description)")
                    return result(
                        FlutterError(code: "\(String(describing: res.error?.code))",
                            message: res.error?.description,
                            details: res.error.debugDescription))
                }
                return result(true)
            }
        } else {
            SwiftTwilioChatPlugin.chatClient?.channelsList()?.channel(withSidOrUniqueName: channelSID){
                res, channel in
                guard (res.isSuccessful()) else {
                    SwiftTwilioChatPlugin.debug("Failed -> \(res.description)")
                    return result(
                        FlutterError(code: "\(String(describing: res.error?.code))",
                            message: res.error?.description,
                            details: res.error.debugDescription))
                }
                channel?.members?.invite(byIdentity: identity){
                    inviteRes in
                    guard (inviteRes.isSuccessful()) else {
                        SwiftTwilioChatPlugin.debug("Failed -> \(inviteRes.description)")
                        return result(
                            FlutterError(code: "\(String(describing: inviteRes.error?.code))",
                                message: inviteRes.error?.description,
                                details: inviteRes.error.debugDescription))
                    }
                    return result(true)
                }
            }
        }
    }
    
    private func updateChannelFriendlyName(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        guard let channel = SwiftTwilioChatPlugin.channel else {
            return result(FlutterError(code: CHANNEL_IS_NULL_CODE, message: CHANNEL_IS_NULL_MESSAGE, details: nil))
        }
        guard let arguments = call.arguments as? [String: Any?] else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'arguments' parameters", details: nil))
        }
        
        guard let friendlyName = arguments["friendlyName"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'friendlyName' parameter", details: nil))
        }
        
        guard let channelSID = arguments["channelSID"] as? String else {
            return result(FlutterError(code: "MISSING_PARAMS", message: "Missing 'channelSID' parameter", details: nil))
        }
        
        if(channelSID == channel.sid){
            channel.setFriendlyName(friendlyName){
                res in
                guard (res.isSuccessful()) else {
                    SwiftTwilioChatPlugin.debug("Failed -> \(res.description)")
                    return result(
                        FlutterError(code: "\(String(describing: res.error?.code))",
                            message: res.error?.description,
                            details: res.error.debugDescription))
                }
                return result(true)
                
            }
        } else {
            SwiftTwilioChatPlugin.chatClient?.channelsList()?.channel(withSidOrUniqueName: channelSID){
                res, channel in
                guard (res.isSuccessful()) else {
                    SwiftTwilioChatPlugin.debug("Failed -> \(res.description)")
                    return result(
                        FlutterError(code: "\(String(describing: res.error?.code))",
                            message: res.error?.description,
                            details: res.error.debugDescription))
                }
                channel!.setFriendlyName(friendlyName){
                    setRes in
                    guard (setRes.isSuccessful()) else {
                        SwiftTwilioChatPlugin.debug("Failed -> \(setRes.description)")
                        return result(
                            FlutterError(code: "\(String(describing: setRes.error?.code))",
                                message: setRes.error?.description,
                                details: setRes.error.debugDescription))
                    }
                    return result(true)
                    
                }
            }
        }
        
        
    }
    
    static func populateChannels(){
        SwiftTwilioChatPlugin.publicChannelDescriptors = NSMutableOrderedSet()
        SwiftTwilioChatPlugin.userChannelDescriptors = NSMutableOrderedSet()
        
        SwiftTwilioChatPlugin.channelsList?.userChannelDescriptors { result, paginator in
            SwiftTwilioChatPlugin.userChannelDescriptors?.addObjects(from: paginator!.items())
        }
        
        SwiftTwilioChatPlugin.channelsList?.publicChannelDescriptors { result, paginator in
            SwiftTwilioChatPlugin.publicChannelDescriptors?.addObjects(from: paginator!.items())
        }
    }
    
}
